%-----------------------------------------------------------------------------%
% Copyright (C) 1996-1997, 1999-2000, 2003 The University of Melbourne.
% This file may only be copied under the terms of the GNU Library General
% Public License - see the file COPYING.LIB in the Mercury distribution.
%-----------------------------------------------------------------------------%
%
% cfloat - the interface to the CLP(R) constraint solver. 
% This module defines the type `cfloat' (short for "constrained float")
% which is a type that is similar in semantics to `float' except
% that it also supports constraints, using the special inst `any'
% for a variable which may be either free or ground.
%
% Note that the operations which mix cfloats and floats are defined
% in the modules cfloat_float and float_cfloat.
%
% The `any' inst used for constraints is not yet fully
% supported by the Mercury mode system, so in some cases it is
% necessary to use explicit calls to cfloat__init.
%
% Main authors: dgj, fjh.
% Stability: low.
%
%-----------------------------------------------------------------------------%

:- module cfloat.

:- interface.

:- import_module list.

%-----------------------------------------------------------------------------%

:- inst constrained == any.

:- mode co == free >> constrained.		% constrain a free variable
:- mode ca == constrained >> constrained.	% constraint across
						% (add a constraint)
:- mode cg == constrained >> ground.		% constraint grounded

	% This is probably a handy set of modes too...

:- inst list_of_constrained == list(constrained).
:- mode list_co == free >> list_of_constrained.
:- mode list_ca == list_of_constrained >> list_of_constrained.
:- mode list_cg == list_of_constrained >> ground.

%-----------------------------------------------------------------------------%

:- solver type cfloat.

% NOTE: the order of the mode declarations for most of the following preds
% is important, since the mode checker tries the given modes of a pred in the
% order they are declared, and we don't want it to use an implied mode.

	% Initialise a solver variable.
	%
:- pred cfloat__init(cfloat::co) is det.

	% negation
	%
:- func '-'(cfloat) = cfloat.
:- mode '-'(ca) = ca is semidet.
:- mode '-'(co) = ca is det.
:- mode '-'(ca) = co is det.

	% equality
	%
:- pred '=='(cfloat, cfloat).
:- mode '=='(ca, ca) is semidet.
:- mode '=='(co, ca) is det.
:- mode '=='(ca, co) is det.

	% disequality
	%
:- pred \==(cfloat, cfloat).
:- mode \==(ca, ca) is semidet.
:- mode \==(co, ca) is det.
:- mode \==(ca, co) is det.

	% addition
	%
:- func '+'(cfloat, cfloat) = cfloat.
:- mode '+'(ca, ca) = ca is semidet.
:- mode '+'(ca, co) = ca is det.
:- mode '+'(co, ca) = ca is det.
:- mode '+'(ca, ca) = co is det.

	% subtraction
	%
:- func '-'(cfloat, cfloat) = cfloat.
:- mode '-'(ca, ca) = ca is semidet.
:- mode '-'(ca, co) = ca is det.
:- mode '-'(co, ca) = ca is det.
:- mode '-'(ca, ca) = co is det.

	% multiplication
	%
:- func '*'(cfloat, cfloat) = cfloat.
:- mode '*'(ca, ca) = ca is semidet.
:- mode '*'(ca, co) = ca is semidet.	% semidet since eg. 0*X=1 fails
:- mode '*'(co, ca) = ca is semidet.	% semidet since eg. X*0=1 fails
:- mode '*'(ca, ca) = co is det.
:- mode '*'(co, co) = ca is det.
:- mode '*'(ca, co) = co is det.
:- mode '*'(co, ca) = co is det.
:- mode '*'(co, co) = co is det.

	% division
	% X / Y = Z :- X = Y * Z, Y \== 0.
	%
:- func '/'(cfloat, cfloat) = cfloat.
:- mode '/'(ca, ca) = ca is semidet.
:- mode '/'(ca, co) = ca is semidet.	% semidet since eg. 0/X=1 fails
:- mode '/'(co, ca) = ca is semidet.	% semidet since eg. X/0=1 fails
:- mode '/'(ca, ca) = co is semidet.	% semidet since eg. 1/0=Y fails
:- mode '/'(co, co) = ca is det.
:- mode '/'(ca, co) = co is semidet.	% XXX really det
:- mode '/'(co, ca) = co is semidet.	% semidet since eg. X/0=Y fails
:- mode '/'(co, co) = co is det.

	% X > Y
	%
:- pred '>'(cfloat, cfloat).
:- mode '>'(ca, ca) is semidet.
:- mode '>'(co, ca) is det.
:- mode '>'(ca, co) is det.
:- mode '>'(co, co) is det.

	% X >= Y
	%
:- pred '>='(cfloat, cfloat).
:- mode '>='(ca, ca) is semidet.
:- mode '>='(co, ca) is det.
:- mode '>='(ca, co) is det.
:- mode '>='(co, co) is det.

	% X < Y
	%
:- pred '<'(cfloat, cfloat).
:- mode '<'(ca, ca) is semidet.
:- mode '<'(ca, co) is det.
:- mode '<'(co, ca) is det.
:- mode '<'(co, co) is det.

	% X =< Y
	%
:- pred '=<'(cfloat, cfloat).
:- mode '=<'(ca, ca) is semidet.
:- mode '=<'(ca, co) is det.
:- mode '=<'(co, ca) is det.
:- mode '=<'(co, co) is det.

%-----------------------------------------------------------------------------%

	% min(X, Y) = (if X < Y then X else Y).
	% Operationally: generally delays until X and Y are ground.
	%
:- func min(cfloat, cfloat) = cfloat.
:- mode min(ca, ca) = ca is semidet.
:- mode min(co, ca) = ca is semidet.
:- mode min(ca, co) = ca is semidet.
:- mode min(ca, ca) = co is det.
:- mode min(ca, co) = co is det.
:- mode min(co, ca) = co is det.
:- mode min(co, co) = ca is det.
:- mode min(co, co) = co is det.

	% max(X, Y) = (if X > Y then X else Y).
	% Operationally: generally delays until X and Y are ground.
	%
:- func max(cfloat, cfloat) = cfloat.
:- mode max(ca, ca) = ca is semidet.
:- mode max(co, ca) = ca is semidet.
:- mode max(ca, co) = ca is semidet.
:- mode max(ca, ca) = co is det.
:- mode max(ca, co) = co is det.
:- mode max(co, ca) = co is det.
:- mode max(co, co) = ca is det.
:- mode max(co, co) = co is det.

	% abs(X) = max(X, -X).
	% Operationally: generally delays until X is ground.
	%
:- func abs(cfloat) = cfloat.
:- mode abs(ca) = ca is semidet.
:- mode abs(co) = ca is semidet.
:- mode abs(ca) = co is det.
:- mode abs(co) = co is det.

	% The usual mathematical sine function.
	% Operationally: generally delays until argument is ground.
	%
:- func sin(cfloat) = cfloat.
:- mode sin(ca) = ca is semidet.
:- mode sin(co) = ca is semidet.
:- mode sin(ca) = co is det.
:- mode sin(co) = co is det.

	% The usual mathematical cosine function.
	% Operationally: generally delays until argument is ground.
	%
:- func cos(cfloat) = cfloat.
:- mode cos(ca) = ca is semidet.
:- mode cos(co) = ca is semidet.
:- mode cos(ca) = co is det.
:- mode cos(co) = co is det.

	% The inverse of the sin function restricted to [-1,1] -> [-pi,].
	% Reports a runtime error if the argument or result is out of range.
	% Generally delays until either argument or result is ground.
	%
:- func arcsin(cfloat) = cfloat.
:- mode arcsin(ca) = ca is semidet.
:- mode arcsin(co) = ca is semidet.
:- mode arcsin(ca) = co is det.
:- mode arcsin(co) = co is det.

	% The inverse of the sin function restricted to [-1,1] -> [-pi,].
	% Reports a runtime error if the argument or result is out of range.
	% Generally delays until either argument or result is ground.
	%
:- func arccos(cfloat) = cfloat.
:- mode arccos(ca) = ca is semidet.
:- mode arccos(co) = ca is semidet.
:- mode arccos(ca) = co is det.
:- mode arccos(co) = co is det.

%-----------------------------------------------------------------------------%

	% Declaratively, cfloat__get_val(X, Y) is true iff Y and X have the
	% same value.  Operationally, cfloat__get_val(X, Y) succeeds only if
	% Y is the only value which the solver variable X can take,
	% given the constraints on X; if there is no unique value, then
	% the predicate will abort at runtime.
	% XXX should this be `cc_multi' rather than `det'?
	%
:- pred cfloat__get_val(cfloat, float).
:- mode cfloat__get_val(ca, out) is det.

%-----------------------------------------------------------------------------%

% The following predicates are used to implement the operators
% defined above.

	% X = Y
:- pred cfloat__eq(cfloat, cfloat).
:- mode cfloat__eq(ca, ca) is semidet.
:- mode cfloat__eq(co, ca) is det.
:- mode cfloat__eq(ca, co) is det.

	% X \= Y
:- pred cfloat__diseq(cfloat, cfloat).
:- mode cfloat__diseq(ca, ca) is semidet.
:- mode cfloat__diseq(co, ca) is det.
:- mode cfloat__diseq(ca, co) is det.

	% cfloat__plus(X, Y, Z) is true iff X+Y=Z
:- pred cfloat__plus(cfloat, cfloat, cfloat).
:- mode cfloat__plus(ca, ca, ca) is semidet.
:- mode cfloat__plus(ca, co, ca) is det.
:- mode cfloat__plus(co, ca, ca) is det.
:- mode cfloat__plus(ca, ca, co) is det.

	% cfloat__minus(X, Y, Z) is true iff X-Y=Z
:- pred cfloat__minus(cfloat, cfloat, cfloat).
:- mode cfloat__minus(ca, ca, ca) is semidet.
:- mode cfloat__minus(ca, co, ca) is det.
:- mode cfloat__minus(co, ca, ca) is det.
:- mode cfloat__minus(ca, ca, co) is det.

	% X*Y=Z
:- pred cfloat__mult(cfloat, cfloat, cfloat).
:- mode cfloat__mult(ca, ca, ca) is semidet.
:- mode cfloat__mult(ca, co, ca) is semidet.	% semidet since eg. 0*X=1 fails
:- mode cfloat__mult(co, ca, ca) is semidet.	% semidet since eg. X*0=1 fails
:- mode cfloat__mult(ca, ca, co) is det.
:- mode cfloat__mult(co, co, ca) is det.
:- mode cfloat__mult(ca, co, co) is det.
:- mode cfloat__mult(co, ca, co) is det.
:- mode cfloat__mult(co, co, co) is det.

%-----------------------------------------------------------------------------%

% The following predicates are used to implement the operators
% defined in cfloat_float.m.

	% X = Y
:- pred cfloat__eq_float(cfloat, float).
:- mode cfloat__eq_float(ca, in) is semidet.
:- mode cfloat__eq_float(co, in) is det.

	% X \= Y
:- pred cfloat__diseq_float(cfloat, float).
:- mode cfloat__diseq_float(ca, in) is semidet.
:- mode cfloat__diseq_float(co, in) is det.

	% X+Y=Z
:- pred cfloat__plus_float(cfloat, float, cfloat).
:- mode cfloat__plus_float(ca, in, ca) is semidet.
:- mode cfloat__plus_float(co, in, ca) is det.
:- mode cfloat__plus_float(ca, in, co) is det.

	% X-Y=Z
:- pred cfloat__minus_float(cfloat, float, cfloat).
:- mode cfloat__minus_float(ca, in, ca) is semidet.
:- mode cfloat__minus_float(co, in, ca) is det.
:- mode cfloat__minus_float(ca, in, co) is det.

	% X*Y=Z
:- pred cfloat__mult_float(cfloat, float, cfloat).
:- mode cfloat__mult_float(ca, in, ca) is semidet.
:- mode cfloat__mult_float(co, in, ca) is semidet. % since eg. X*0=1 fails
:- mode cfloat__mult_float(ca, in, co) is det.
:- mode cfloat__mult_float(co, in, co) is det.

	% X/Y = Z
	% reports a runtime error if Y = 0.
:- pred cfloat__div_float(cfloat, float, cfloat).
:- mode cfloat__div_float(ca, in, ca) is semidet.
:- mode cfloat__div_float(co, in, ca) is det.
:- mode cfloat__div_float(ca, in, co) is det.
:- mode cfloat__div_float(co, in, co) is det.

	% X > Y
:- pred cfloat__gt_float(cfloat, float).
:- mode cfloat__gt_float(ca, in) is semidet.
:- mode cfloat__gt_float(co, in) is det.

	% X >= Y
:- pred cfloat__ge_float(cfloat, float).
:- mode cfloat__ge_float(ca, in) is semidet.
:- mode cfloat__ge_float(co, in) is det.

	% X < Y
:- pred cfloat__lt_float(cfloat, float).
:- mode cfloat__lt_float(ca, in) is semidet.
:- mode cfloat__lt_float(co, in) is det.

	% X =< Y
:- pred cfloat__le_float(cfloat, float).
:- mode cfloat__le_float(ca, in) is semidet.
:- mode cfloat__le_float(co, in) is det.

	% The compiler automatically generates calls to this
	% predicate to initialize variables to inst `any'.
:- pred cfloat__cfloat_init_any(cfloat).
:- mode cfloat__cfloat_init_any(co) is det.

%----------------------------------------------------------------------------%
%----------------------------------------------------------------------------%

:- implementation.
:- import_module float, cfloat_float, float_cfloat.

%----------------------------------------------------------------------------%

X == Y :- cfloat__eq(X, Y).
X \== Y :- cfloat__diseq(X, Y).
- X = 0.0 - X.
X + Y = Z :- cfloat__plus(X, Y, Z).
X - Y = Z :- cfloat__minus(X, Y, Z).
X * Y = Z :- cfloat__mult(X, Y, Z).
X / Y = Z :- Y \== 0.0, X = Y * Z.

cfloat__cfloat_init_any(Svar) :- cfloat__init(Svar).

%----------------------------------------------------------------------------%

:- solver type cfloat
	where 	representation 	is int,
		initialisation 	is cfloat__cfloat_init_any,
		ground		is ground,
		any		is ground,
		equality 	is cfloat__eq.		

:- pragma foreign_decl("C", "

#include ""mercury_conf.h""
#include ""mercury_trail.h""
#include ""clpr/emul.h""
#include ""clpr/solver.h""
#include ""clpr/clpr_misc.h""
#include <stdio.h>

#ifndef MR_USE_TRAIL
  #error This module may only be used in a trailed grade (*.tr).
#endif

/*
** There are three data structures involved in trailing:
**	- the Mercury trail
**	- the cfloat.m choice point stack
**	- the CLP(R) trail
** The Mercury trail contains function trail entries which point to
** the cfloat.m choice point stack.
** The cfloat.m choice point stack is represented a singly-linked list,
** with a free list used to minimize the cost of memory allocation.
** Its entries store the values of the CLP(R) global variables, including
** the `CLPR_trtop' variable, which points to the top of the CLP(R) trail.
*/
#ifndef ML_CFLOAT_CHOICEPOINT_GUARD
#define ML_CFLOAT_CHOICEPOINT_GUARD
typedef struct ML_cfloat_choicepoint {
	int stamp;
	int slack_id;
	int solver_id;
	int trtop;
	NL_EQN_ptr nl_eqn_top;
	struct ML_cfloat_choicepoint *next;
} ML_cfloat_choicepoint;
#endif

#define ML_cfloat_maybe_trail_solver()				\\
	( CLPR_stamp != MR_current_choicepoint_id() ? (		\\
			ML_cfloat_trail_solver(),		\\
			(void)0					\\
	) : ( 							\\
			(void)0					\\
	) )

void ML_cfloat_trail_solver(void);
void ML_cfloat_untrail_func(void *datum, MR_untrail_reason reason);

").

:- pragma foreign_code("C", "

static ML_cfloat_choicepoint ML_cfloat_first_choicepoint;
static ML_cfloat_choicepoint * ML_cfloat_current_cp =
					&ML_cfloat_first_choicepoint;

void
ML_cfloat_trail_solver(void)
{
	ML_cfloat_choicepoint *choicepoint = ML_cfloat_current_cp;
	choicepoint->stamp = CLPR_stamp;
	choicepoint->slack_id = CLPR_slack_id;
	choicepoint->solver_id = CLPR_solver_id;
	choicepoint->trtop = CLPR_trtop;
	choicepoint->nl_eqn_top = CLPR_nl_eqn_top;
	MR_trail_function(ML_cfloat_untrail_func, choicepoint);

	if (choicepoint->next == NULL) {
		ML_cfloat_choicepoint *next_choicepoint;
		next_choicepoint = MR_GC_NEW(ML_cfloat_choicepoint);
		next_choicepoint->next = NULL;
		choicepoint->next = next_choicepoint;
	}
	ML_cfloat_current_cp = choicepoint->next;

	CLPR_stamp = MR_current_choicepoint_id();
}

static void
ML_cfloat_check_floundering(NL_EQN_ptr old_top)
{
	NL_EQN_ptr ptr;
	bool floundered;

	/*
	** Check if we have added any non-linear equations
	** since `old_top' and not yet woken them up.
	*/

	floundered = FALSE;
	for (ptr = CLPR_nl_eqn_top; ptr != old_top; ptr = ptr->next) {
		if (ptr->type != WOKEN && ptr->partner == NULL) {
			floundered = TRUE;
		}
	}

	if (floundered) {
		fflush(stdout);
		fprintf(stderr,
			""\\ncfloat.m: warning: goal floundered.\\n""
			""\\tThis solver can only solve linear constraints.\\n""
			""\\tThe following delayed goals\\n""
		);
		for (ptr = CLPR_nl_eqn_top; ptr != old_top; ptr = ptr->next) {
			if (ptr->type != WOKEN && ptr->partner == NULL) {
				fprintf(stderr, ""\\t\\t"");
				CLPR_print_nl_eqn(ptr);
			}
		}
		fprintf(stderr, ""\\twere assumed to be solvable.\\n"");
	}
}

void
ML_cfloat_untrail_func(void *datum, MR_untrail_reason reason)
{
	ML_cfloat_choicepoint *choicepoint = (ML_cfloat_choicepoint *) datum;
	switch (reason) {
		case MR_undo:
		case MR_exception:
		case MR_retry:
		{
			int old_trtop, i;

			CLPR_stamp = choicepoint->stamp;
			CLPR_slack_id = choicepoint->slack_id;
			CLPR_solver_id = choicepoint->solver_id;
			old_trtop = choicepoint->trtop;
			for (i = CLPR_trtop - 1; i >= old_trtop; i--) {
				CLPR_solver_backtrack((MR_Word) CLPR_trail[i]);
			}
			CLPR_trtop = old_trtop;
			ML_cfloat_current_cp = choicepoint;
			break;
		}
		case MR_commit:
			CLPR_stamp = choicepoint->stamp;
			/* fall through */
		case MR_solve:
			ML_cfloat_check_floundering(choicepoint->nl_eqn_top);
			break;
		default:
			MR_fatal_error(""cfloat.m: unknown MR_untrail_reason"");
	}
}
").

:- pragma foreign_decl("C", "

#ifndef ML_CFLOAT_HEADER_GUARD
#define ML_CFLOAT_HEADER_GUARD

typedef MR_Integer ML_svar;

/*
** The following ""functions"" are currently all implemented using macros.
** (Though for many it might actually be better if they were functions...)
** The declarations below are just documentation.
*/

bool ML_cfloat_eq(ML_svar, ML_svar);
bool ML_cfloat_eq_float(ML_svar, MR_Float);
/* bool ML_cfloat_diseq(ML_svar, ML_svar, bool&); */
/* bool ML_cfloat_diseq_float(ML_svar, MR_Float, bool&); */
bool ML_cfloat_gt(ML_svar, ML_svar);
bool ML_cfloat_gt_float(ML_svar, MR_Float);
bool ML_cfloat_lt(ML_svar, ML_svar);
bool ML_cfloat_lt_float(ML_svar, MR_Float);
bool ML_cfloat_ge(ML_svar, ML_svar);
bool ML_cfloat_ge_float(ML_svar, MR_Float);
bool ML_cfloat_le(ML_svar, ML_svar);
bool ML_cfloat_le_float(ML_svar, MR_Float);
bool ML_cfloat_plus(ML_svar, ML_svar, ML_svar);
bool ML_cfloat_plus_float(ML_svar, MR_Float, ML_svar);
bool ML_cfloat_minus(ML_svar, ML_svar, ML_svar);
bool ML_cfloat_minus_float(ML_svar, MR_Float, ML_svar);
bool ML_cfloat_mult(ML_svar, ML_svar, ML_svar);
bool ML_cfloat_mult_float(ML_svar, MR_Float, ML_svar);
bool ML_cfloat_div_float(ML_svar, MR_Float, ML_svar);
bool ML_cfloat_min(ML_svar, ML_svar, ML_svar);
bool ML_cfloat_max(ML_svar, ML_svar, ML_svar);
bool ML_cfloat_abs(ML_svar, ML_svar);
bool ML_cfloat_sin(ML_svar, ML_svar);
bool ML_cfloat_cos(ML_svar, ML_svar);
bool ML_cfloat_arcsin(ML_svar, ML_svar);
bool ML_cfloat_arccos(ML_svar, ML_svar);

/*
** create a fresh cfloat variable (of inst `any')
*/
#define ML_cfloat_init_solver_var(x) \\
	((x) = ++CLPR_solver_id)

#if 1
/*
** Svar1 = Svar2 :-
**	1.0*Svar1 - 1.0*Svar2 = 0.0.
*/
#define ML_cfloat_eq(Svar1, Svar2)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(1.0, Svar1),		\\
		CLPR_s_evar(-1.0, Svar2),		\\
		/* return */ CLPR_s_econ(0.0)		\\
	)

/*
** Svar1 = Val :-
**	-1.0*Svar1 = Val.
*/
#define ML_cfloat_eq_float(Svar1, Val)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(-1.0, Svar),		\\
		/* return */ CLPR_s_econ(Val)		\\
	)

/*
** Svar1 \\= Svar2 :-
**	Svar1 - Svar2 = Diff,
**	abs(Diff) = AbsDiff,
**	AbsDiff > 0.
*/
#define ML_cfloat_diseq(Svar1, Svar2, result)			\\
	do {							\\
		MR_Word Diff, AbsDiff;				\\
								\\
		ML_cfloat_init_solver_var(Diff);		\\
		ML_cfloat_init_solver_var(AbsDiff);		\\
		(void) ML_cfloat_minus((Svar1), (Svar2), Diff);	\\
		(void) ML_cfloat_abs(Diff, AbsDiff);		\\
		(result) = ML_cfloat_gt_float(AbsDiff, 0.0);	\\
	} while(0)

/*
** Svar \\= Val :-
**	Svar - Val = Diff,
**	abs(Diff) = AbsDiff,
**	AbsDiff > 0.
*/
#define ML_cfloat_diseq_float(Svar, Val, result)			\\
	do {								\\
		MR_Word Diff, AbsDiff;					\\
									\\
		ML_cfloat_init_solver_var(Diff);			\\
		ML_cfloat_init_solver_var(AbsDiff);			\\
		(void) ML_cfloat_minus_float((Svar), (Val), Diff);	\\
		(void) ML_cfloat_abs(Diff, AbsDiff);			\\
		(result) = ML_cfloat_gt_float(AbsDiff, 0.0);		\\
	} while(0)

/*
** Svar1 + Svar2 = Svar3 :-
**	1.0*Svar1 + 1.0*Svar2 + -1.0*Svar3 = 0.0.
*/
#define ML_cfloat_plus(Svar1, Svar2, Svar3)		\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(1.0, Svar1),		\\
		CLPR_s_evar(1.0, Svar2),		\\
		CLPR_s_evar(-1.0, Svar3),		\\
		/* return */ CLPR_s_econ(0.0)		\\
	)

/*
** Svar1 + Val = Svar2 :-
**	1.0*Svar1 + Val + -1.0*Svar2 = 0.0.
*/
#define ML_cfloat_plus_float(Svar1, Val, Svar2)			\\
	(							\\
		ML_cfloat_maybe_trail_solver(),			\\
		CLPR_init_eqn(),				\\
		CLPR_s_evar(1.0, Svar1),			\\
		CLPR_eqn_scratch->constant += (double) Val,	\\
		CLPR_s_evar(-1.0, Svar2),			\\
		/* return */ CLPR_s_econ(0.0)			\\
	)

/*
** Svar1 - Svar2 = Svar3 :-
**	1.0*Svar1 + -1.0*Svar2 + -1.0*Svar3 = 0.0.
*/
#define ML_cfloat_minus(Svar1, Svar2, Svar3)		\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(1.0, Svar1),		\\
		CLPR_s_evar(-1.0, Svar2),		\\
		CLPR_s_evar(-1.0, Svar3),		\\
		/* return */ CLPR_s_econ(0.0)		\\
	)

/*
** Svar1 - Val = Svar2 :-
**	1.0*Svar1 - Val + -1.0*Svar2 = 0.0.
*/
#define ML_cfloat_minus_float(Svar1, Val, Svar2)		\\
	(							\\
		ML_cfloat_maybe_trail_solver(),			\\
		CLPR_init_eqn(),				\\
		CLPR_s_evar(1.0, Svar1),			\\
		CLPR_eqn_scratch->constant -= (double) Val,	\\
		CLPR_s_evar(-1.0, Svar2),			\\
		/* return */ CLPR_s_econ(0.0)			\\
	)

/*
** Svar1 * Svar2 = Svar3 :-
**	1.0*Svar3 = Svar1 * Svar2.
*/
#define ML_cfloat_mult(Svar1, Svar2, Svar3)				\\
	(								\\
		ML_cfloat_maybe_trail_solver(),				\\
		/* CLPR_init_eqn() not needed */			\\
		/* return */ CLPR_s_mult(1.0, Svar3, Svar2, Svar1)	\\
	)

/*
** Svar1 * Val = Svar2 :-
**	Val*Svar1 + -1.0*Svar2 = 0.0.
*/
#define ML_cfloat_mult_float(Svar1, Val, Svar2)		\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar((double) Val, Svar1),	\\
		CLPR_s_evar(-1.0, Svar2),		\\
		/* return */ CLPR_s_econ(0.0)		\\
	)

/*
** Svar1 / Val = Svar2 :-
**	( Val = 0 ->
**		error(""division by zero"")
**	;
**		(1.0/Val)*Svar1 + -1.0*Svar2 = 0.0
**	).
** Note that we need to explicitly check for division by zero,
** because on some platforms the call to 1.0/Val will just return
** infinity, rather than a run-time error (floating point exception).
*/
#define ML_cfloat_div_float(Svar1, Val, Svar2)		\\
	( Val == 0 ? (					\\
		MR_fatal_error(""division by zero""),	\\
		FALSE					\\
	) : (						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(1.0 / (double) Val, Svar1),	\\
		CLPR_s_evar(-1.0, Svar2),		\\
		/* return */ CLPR_s_econ(0.0)		\\
	) )

/*
** Svar1 > Svar2 :-
**	1.0*Svar1 + -1.0*Svar2 > 0.0.
*/
#define ML_cfloat_gt(Svar1, Svar2)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(1.0, Svar1),		\\
		CLPR_s_evar(-1.0, Svar2),		\\
		/* return */ CLPR_s_gt0(0.0)		\\
	)

/*
** Svar1 >= Svar2 :-
**	1.0*Svar1 + -1.0*Svar2 >= 0.0.
*/
#define ML_cfloat_ge(Svar1, Svar2)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(1.0, Svar1),		\\
		CLPR_s_evar(-1.0, Svar2),		\\
		/* return */ CLPR_s_ge0(0.0)		\\
	)

/*
** Svar1 < Svar2 :-
**	-1.0*Svar1 + 1.0*Svar2 > 0.0.
*/
#define ML_cfloat_lt(Svar1, Svar2)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(-1.0, Svar1),		\\
		CLPR_s_evar(1.0, Svar2),		\\
		/* return */ CLPR_s_gt0(0.0)		\\
	)

/*
** Svar1 =< Svar2 :-
**	-1.0*Svar1 + 1.0*Svar2 >= 0.0.
*/
#define ML_cfloat_le(Svar1, Svar2)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(-1.0, Svar1),		\\
		CLPR_s_evar(1.0, Svar2),		\\
		/* return */ CLPR_s_ge0(0.0)		\\
	)

/*
** Svar1 > Val :-
**	1.0*Svar1 > Val.
*/
#define ML_cfloat_gt_float(Svar, Val)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(1.0, Svar),			\\
		/* return */ CLPR_s_gt0((double) Val)	\\
	)

/*
** Svar >= Val :-
**	1.0*Svar >= Val.
*/
#define ML_cfloat_ge_float(Svar, Val)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(1.0, Svar),			\\
		/* return */ CLPR_s_ge0((double) Val)	\\
	)


/*
** Svar1 < Val :-
**	-1.0*Svar1 > -Val.
*/
#define ML_cfloat_lt_float(Svar1, Val)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(-1.0, Svar),		\\
		/* return */ CLPR_s_gt0((double) -Val)	\\
	)

/*
** Svar >= Val :-
**	-1.0*Svar >= -Val.
*/
#define ML_cfloat_le_float(Svar1, Val)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		CLPR_init_eqn(),			\\
		CLPR_s_evar(-1.0, Svar),		\\
		/* return */ CLPR_s_ge0((double) -Val)	\\
	)

/*
** min(Svar1, Svar2) = Svar3
*/
#define ML_cfloat_min(Svar1, Svar2, Svar3)			\\
	(							\\
		ML_cfloat_maybe_trail_solver(),			\\
		/* return */ CLPR_s_min(Svar1, Svar2, Svar3)	\\
	)

/*
** max(Svar1, Svar2) = Svar3
*/
#define ML_cfloat_max(Svar1, Svar2, Svar3)			\\
	(							\\
		ML_cfloat_maybe_trail_solver(),			\\
		/* return */ CLPR_s_max(Svar3, Svar1, Svar2)	\\
	)

/*
** abs(Svar1) = Svar2
*/
#define ML_cfloat_abs(Svar1, Svar2)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		/* return */ CLPR_s_abs(Svar2, Svar1)	\\
	)

/*
** sin(Svar1) = Svar2
*/
#define ML_cfloat_sin(Svar1, Svar2)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		/* return */ CLPR_s_sin(Svar2, Svar1)	\\
	)

/*
** cos(Svar1) = Svar2
*/
#define ML_cfloat_cos(Svar1, Svar2)			\\
	(						\\
		ML_cfloat_maybe_trail_solver(),		\\
		/* return */ CLPR_s_cos(Svar2, Svar1)	\\
	)

/*
** arcsin(Svar1) = Svar2
*/
#define ML_cfloat_arcsin(Svar1, Svar2)				\\
	(							\\
		ML_cfloat_maybe_trail_solver(),			\\
		/* return */ CLPR_s_arcsin(Svar2, Svar1)	\\
	)

/*
** arccos(Svar1) = Svar2
*/
#define ML_cfloat_arccos(Svar1, Svar2)				\\
	(							\\
		ML_cfloat_maybe_trail_solver(),			\\
		/* return */ CLPR_s_arccos(Svar2, Svar1)	\\
	)

#endif

#endif /* ML_CFLOAT_HEADER_GUARD */

").

:- pragma foreign_code("C", "
/*
INIT init_cfloat_module
*/
/*
** We need to ensure that CLP(R) engine gets initialized.
** But we can't rely on the special INIT comment above,
** since initialization functions registered using that mechanism
** only get called in certain grades.
** So we use GCC's `constructor' attribute.
*/
#ifdef __GNUC__
  void init_cfloat_module(void) __attribute__((constructor)); /* XXX gcc-ism */
#else
  #error The CLP(R) interface requires GNU C
#endif
void init_cfloat_module(void) {
	CLPR_init();
}

").

:- pragma c_code(cfloat__init(Svar::co), "ML_cfloat_init_solver_var(Svar);").

:- pragma c_code(cfloat__plus(Svar1::ca, Svar2::ca, Svar3::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_plus(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__plus(Svar1::ca, Svar2::ca, Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_plus(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__plus(Svar1::ca, Svar2::co, Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_plus(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__plus(Svar1::co, Svar2::ca, Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		(void) ML_cfloat_plus(Svar1, Svar2, Svar3);
	").

:- pragma c_code(cfloat__plus_float(Svar1::ca, Val::in, Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_plus_float(Svar1, Val, Svar2);
	").
:- pragma c_code(cfloat__plus_float(Svar1::co, Val::in, Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		(void) ML_cfloat_plus_float(Svar1, Val, Svar2);
	").
:- pragma c_code(cfloat__plus_float(Svar1::ca, Val::in, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_plus_float(Svar1, Val, Svar2);
	").


:- pragma c_code(cfloat__minus(Svar1::ca, Svar2::ca, Svar3::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_minus(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__minus(Svar1::ca, Svar2::ca, Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_minus(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__minus(Svar1::ca, Svar2::co, Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_minus(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__minus(Svar1::co, Svar2::ca, Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		(void) ML_cfloat_minus(Svar1, Svar2, Svar3);
	").

:- pragma c_code(cfloat__minus_float(Svar1::ca, Val::in, Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_minus_float(Svar1, Val, Svar2);
	").
:- pragma c_code(cfloat__minus_float(Svar1::co, Val::in, Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		(void) ML_cfloat_minus_float(Svar1, Val, Svar2);
	").
:- pragma c_code(cfloat__minus_float(Svar1::ca, Val::in, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_minus_float(Svar1, Val, Svar2);
	").


:- pragma c_code(cfloat__mult(Svar1::ca, Svar2::ca, Svar3::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_mult(Svar1, Svar2, Svar3);
	").	
:- pragma c_code(cfloat__mult(Svar1::ca, Svar2::co, Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar2);
		SUCCESS_INDICATOR = ML_cfloat_mult(Svar1, Svar2, Svar3);
	").	
:- pragma c_code(cfloat__mult(Svar1::co, Svar2::ca, Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		SUCCESS_INDICATOR = ML_cfloat_mult(Svar1, Svar2, Svar3);
	").	
:- pragma c_code(cfloat__mult(Svar1::ca, Svar2::ca, Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_mult(Svar1, Svar2, Svar3);
	").	
:- pragma c_code(cfloat__mult(Svar1::co, Svar2::co, Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_mult(Svar1, Svar2, Svar3);
	").	
:- pragma c_code(cfloat__mult(Svar1::co, Svar2::ca, Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_mult(Svar1, Svar2, Svar3);
	").	
:- pragma c_code(cfloat__mult(Svar1::ca, Svar2::co, Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_mult(Svar1, Svar2, Svar3);
	").	
:- pragma c_code(cfloat__mult(Svar1::co, Svar2::co, Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_mult(Svar1, Svar2, Svar3);
	").	

:- pragma c_code(cfloat__mult_float(Svar1::ca, Val::in, Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_mult_float(Svar1, Val, Svar2);
	").
:- pragma c_code(cfloat__mult_float(Svar1::co, Val::in, Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		SUCCESS_INDICATOR = ML_cfloat_mult_float(Svar1, Val, Svar2);
	").
:- pragma c_code(cfloat__mult_float(Svar1::ca, Val::in, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_mult_float(Svar1, Val, Svar2);
	").
:- pragma c_code(cfloat__mult_float(Svar1::co, Val::in, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_mult_float(Svar1, Val, Svar2);
	").


:- pragma c_code(cfloat__div_float(Svar1::ca, Val::in, Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_div_float(Svar1, Val, Svar2);
	").
:- pragma c_code(cfloat__div_float(Svar1::co, Val::in, Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		(void) ML_cfloat_div_float(Svar1, Val, Svar2);
	").
:- pragma c_code(cfloat__div_float(Svar1::ca, Val::in, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_div_float(Svar1, Val, Svar2);
	").
:- pragma c_code(cfloat__div_float(Svar1::co, Val::in, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_div_float(Svar1, Val, Svar2);
	").

%-----------------------------------------------------------------------------%

:- pragma c_code(cfloat__eq(Svar1::ca, Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_eq(Svar1, Svar2);
	").
:- pragma c_code(cfloat__eq(Svar1::co, Svar2::ca),
	"
		Svar1 = Svar2;
	").
:- pragma c_code(cfloat__eq(Svar1::ca, Svar2::co),
	"
		Svar2 = Svar1;
	").

:- pragma c_code(cfloat__eq_float(Svar::ca, Val::in),
	"
		SUCCESS_INDICATOR = ML_cfloat_eq_float(Svar, Val);
	").
:- pragma c_code(cfloat__eq_float(Svar::co, Val::in),
	"
		ML_cfloat_init_solver_var(Svar);
		(void) ML_cfloat_eq_float(Svar, Val);
	").


:- pragma c_code(cfloat__diseq(Svar1::ca, Svar2::ca),
	"{
		ML_cfloat_diseq(Svar1, Svar2, SUCCESS_INDICATOR);
	}").
:- pragma c_code(cfloat__diseq(Svar1::co, Svar2::ca),
	"{
		bool result;
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_diseq(Svar1, Svar2, result);
		(void) result; /* result not used */
	}").
:- pragma c_code(cfloat__diseq(Svar1::ca, Svar2::co),
	"{
		bool result;
		ML_cfloat_init_solver_var(Svar2);
		ML_cfloat_diseq(Svar1, Svar2, result);
		(void) result; /* result not used */
	}").

:- pragma c_code(cfloat__diseq_float(Svar::ca, Val::in),
	"{
		ML_cfloat_diseq_float(Svar, Val, SUCCESS_INDICATOR);
	}").
:- pragma c_code(cfloat__diseq_float(Svar::co, Val::in),
	"{
		bool result;
		ML_cfloat_init_solver_var(Svar);
		ML_cfloat_diseq_float(Svar, Val, result);
		(void) result; /* result not used */
	}").

%-----------------------------------------------------------------------------%

:- pragma c_code('>'(Svar1::ca, Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_gt(Svar1, Svar2);
	").
:- pragma c_code('>'(Svar1::co, Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		(void) ML_cfloat_gt(Svar1, Svar2);
	").
:- pragma c_code('>'(Svar1::ca, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_gt(Svar1, Svar2);
	").
:- pragma c_code('>'(Svar1::co, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_gt(Svar1, Svar2);
	").

:- pragma c_code('>='(Svar1::ca, Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_ge(Svar1, Svar2);
	").
:- pragma c_code('>='(Svar1::co, Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		(void) ML_cfloat_ge(Svar1, Svar2);
	").
:- pragma c_code('>='(Svar1::ca, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_ge(Svar1, Svar2);
	").
:- pragma c_code('>='(Svar1::co, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_ge(Svar1, Svar2);
	").

:- pragma c_code(cfloat__gt_float(Svar::ca, Val::in),
	"
		SUCCESS_INDICATOR = ML_cfloat_gt_float(Svar, Val);
	").
:- pragma c_code(cfloat__gt_float(Svar::co, Val::in),
	"
		ML_cfloat_init_solver_var(Svar);
		(void) ML_cfloat_gt_float(Svar, Val);
	").

:- pragma c_code(cfloat__ge_float(Svar::ca, Val::in),
	"
		SUCCESS_INDICATOR = ML_cfloat_ge_float(Svar, Val);
	").
:- pragma c_code(cfloat__ge_float(Svar::co, Val::in),
	"
		ML_cfloat_init_solver_var(Svar);
		(void) ML_cfloat_ge_float(Svar, Val);
	").

:- pragma c_code('<'(Svar1::ca, Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_lt(Svar1, Svar2);
	").
:- pragma c_code('<'(Svar1::ca, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_lt(Svar1, Svar2);
	").
:- pragma c_code('<'(Svar1::co, Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		(void) ML_cfloat_lt(Svar1, Svar2);
	").
:- pragma c_code('<'(Svar1::co, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_lt(Svar1, Svar2);
	").
:- pragma c_code('=<'(Svar1::ca, Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_le(Svar1, Svar2);
	").
:- pragma c_code('=<'(Svar1::ca, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_le(Svar1, Svar2);
	").
:- pragma c_code('=<'(Svar1::co, Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		(void) ML_cfloat_le(Svar1, Svar2);
	").
:- pragma c_code('=<'(Svar1::co, Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_le(Svar1, Svar2);
	").

:- pragma c_code(cfloat__lt_float(Svar::ca, Val::in),
	"
		SUCCESS_INDICATOR = ML_cfloat_lt_float(Svar, Val);
	").
:- pragma c_code(cfloat__lt_float(Svar::co, Val::in),
	"
		ML_cfloat_init_solver_var(Svar);
		(void) ML_cfloat_lt_float(Svar, Val);
	").

:- pragma c_code(cfloat__le_float(Svar::ca, Val::in),
	"
		SUCCESS_INDICATOR = ML_cfloat_le_float(Svar, Val);
	").
:- pragma c_code(cfloat__le_float(Svar::co, Val::in),
	"
		ML_cfloat_init_solver_var(Svar);
		(void) ML_cfloat_le_float(Svar, Val);
	").

%-----------------------------------------------------------------------------%

:- pragma c_code(cfloat__get_val(Svar::ca, Val::out),
	"
		if (!CLPR_is_ground(Svar, &Val)) {
			fprintf(stderr, ""ERROR: Var %ld not ground.\\n"",
				(long) Svar);
			MR_fatal_error(""Solver variable not ground ""
				""in cfloat__get_val"");
		}
	").

%-----------------------------------------------------------------------------%

:- pragma c_code(cfloat__min(Svar1::ca, Svar2::ca) = (Svar3::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_min(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__min(Svar1::co, Svar2::ca) = (Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		SUCCESS_INDICATOR = ML_cfloat_min(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__min(Svar1::ca, Svar2::co) = (Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar2);
		SUCCESS_INDICATOR = ML_cfloat_min(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__min(Svar1::ca, Svar2::ca) = (Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_min(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__min(Svar1::co, Svar2::co) = (Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_min(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__min(Svar1::co, Svar2::ca) = (Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_min(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__min(Svar1::ca, Svar2::co) = (Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_min(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__min(Svar1::co, Svar2::co) = (Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_min(Svar1, Svar2, Svar3);
	").

:- pragma c_code(cfloat__max(Svar1::ca, Svar2::ca) = (Svar3::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_max(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__max(Svar1::co, Svar2::ca) = (Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		SUCCESS_INDICATOR = ML_cfloat_max(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__max(Svar1::ca, Svar2::co) = (Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar2);
		SUCCESS_INDICATOR = ML_cfloat_max(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__max(Svar1::ca, Svar2::ca) = (Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_max(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__max(Svar1::co, Svar2::co) = (Svar3::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_max(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__max(Svar1::co, Svar2::ca) = (Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_max(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__max(Svar1::ca, Svar2::co) = (Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_max(Svar1, Svar2, Svar3);
	").
:- pragma c_code(cfloat__max(Svar1::co, Svar2::co) = (Svar3::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		ML_cfloat_init_solver_var(Svar3);
		(void) ML_cfloat_max(Svar1, Svar2, Svar3);
	").

%-----------------------------------------------------------------------------%

:- pragma c_code(cfloat__abs(Svar1::ca) = (Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_abs(Svar1, Svar2);
	").
:- pragma c_code(cfloat__abs(Svar1::co) = (Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		SUCCESS_INDICATOR = ML_cfloat_abs(Svar1, Svar2);
	").
:- pragma c_code(cfloat__abs(Svar1::ca) = (Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_abs(Svar1, Svar2);
	").
:- pragma c_code(cfloat__abs(Svar1::co) = (Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_abs(Svar1, Svar2);
	").

%-----------------------------------------------------------------------------%

:- pragma c_code(cfloat__sin(Svar1::ca) = (Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_sin(Svar1, Svar2);
	").
:- pragma c_code(cfloat__sin(Svar1::co) = (Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		SUCCESS_INDICATOR = ML_cfloat_sin(Svar1, Svar2);
	").
:- pragma c_code(cfloat__sin(Svar1::ca) = (Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_sin(Svar1, Svar2);
	").
:- pragma c_code(cfloat__sin(Svar1::co) = (Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_sin(Svar1, Svar2);
	").

:- pragma c_code(cfloat__cos(Svar1::ca) = (Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_cos(Svar1, Svar2);
	").
:- pragma c_code(cfloat__cos(Svar1::co) = (Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		SUCCESS_INDICATOR = ML_cfloat_cos(Svar1, Svar2);
	").
:- pragma c_code(cfloat__cos(Svar1::ca) = (Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_cos(Svar1, Svar2);
	").
:- pragma c_code(cfloat__cos(Svar1::co) = (Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_cos(Svar1, Svar2);
	").

%-----------------------------------------------------------------------------%

:- pragma c_code(cfloat__arcsin(Svar1::ca) = (Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_arcsin(Svar1, Svar2);
	").
:- pragma c_code(cfloat__arcsin(Svar1::co) = (Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		SUCCESS_INDICATOR = ML_cfloat_arcsin(Svar1, Svar2);
	").
:- pragma c_code(cfloat__arcsin(Svar1::ca) = (Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_arcsin(Svar1, Svar2);
	").
:- pragma c_code(cfloat__arcsin(Svar1::co) = (Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_arcsin(Svar1, Svar2);
	").

:- pragma c_code(cfloat__arccos(Svar1::ca) = (Svar2::ca),
	"
		SUCCESS_INDICATOR = ML_cfloat_arccos(Svar1, Svar2);
	").
:- pragma c_code(cfloat__arccos(Svar1::co) = (Svar2::ca),
	"
		ML_cfloat_init_solver_var(Svar1);
		SUCCESS_INDICATOR = ML_cfloat_arccos(Svar1, Svar2);
	").
:- pragma c_code(cfloat__arccos(Svar1::ca) = (Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_arccos(Svar1, Svar2);
	").
:- pragma c_code(cfloat__arccos(Svar1::co) = (Svar2::co),
	"
		ML_cfloat_init_solver_var(Svar1);
		ML_cfloat_init_solver_var(Svar2);
		(void) ML_cfloat_arccos(Svar1, Svar2);
	").

%-----------------------------------------------------------------------------%
:- end_module cfloat.
%-----------------------------------------------------------------------------%
