%-----------------------------------------------------------------------------%
% Copyright (C) 1996-1997, 2000 The University of Melbourne.
% This file may only be copied under the terms of the GNU Library General
% Public License - see the file COPYING.LIB in the Mercury distribution.
%-----------------------------------------------------------------------------%
%
% dump - This module provides predicates for solving constraint goals
%	 and then writing the constraint relationships to stdout.
%
% Main author: dgj.
% Interface resigned by fjh.
% Stability: low/medium.
%
%-----------------------------------------------------------------------------%

:- module dump.

:- interface.

:- import_module cfloat, list, string, io.

%-----------------------------------------------------------------------------%

	% dump_one_solution(Pred, IO0, IO) is true iff
	%	there is some constraint C between variables in CfloatList
	%	such that Pred(CfloatsList, NameList) is satisfied if C holds,
	%	and IO is obtained from IO0 by writing a representation of C
	%	to stdout, using the names in the corresponding list NamesList;
	%	or, Pred(CfloatList, NameList) has no solution, and IO is
	%	obtained from IO0 by writing "No solutions.\n" to stdout.
	%
:- pred dump_one_solution(pred(list(cfloat), list(string)), io, io).
:- mode dump_one_solution(
		pred(list_co, out) is cc_nondet, di, uo) is cc_multi.
:- mode dump_one_solution(
		pred(list_co, out) is semidet, di, uo) is cc_multi.

	% dump_cfloat(X, IO0, IO) is true iff
	%	IO is obtained from IO0 by writing either
	%		a floating point value F such that X has value F,
	%	or
	%		"_v<N>", where <N> is an integer.
	%
:- pred dump_cfloat(cfloat::ca, io::di, io::uo) is cc_multi.

	% XXX this one is a non-logical hack, use only for debugging
	%
:- impure pred unsafe_dump(list(cfloat)::list_ca, list(string)::in) is det.

	% XXX this one is a non-logical hack, use only for debugging
	%
:- impure pred unsafe_dump_cfloat(cfloat::ca) is det.

	% for debugging only... this pred (non-logically) dumps the CLP(R)
	% tableaus to standard error at runtime.
	%
:- impure pred unsafe_dump_tableaus is det.

%-----------------------------------------------------------------------------%
%-----------------------------------------------------------------------------%

:- implementation.
:- import_module int, require.

:- pragma foreign_decl("C", "#include ""clpr/solver.h""").
:- pragma foreign_decl("C", "#include ""clpr/clpr_misc.h""").

:- pragma foreign_proc("C",
	dump_cfloat(Svar::ca, IO0::di, IO::uo),
	[promise_pure, will_not_call_mercury],
"{
	double val;
	if (CLPR_is_ground(Svar, &val)) {
		printf(""%f"", val);
	} else {
		printf(""_v%ld"", (long) Svar);
	}
	IO = IO0;
}").


:- pragma foreign_proc("C",
	unsafe_dump_cfloat(Svar::ca),
	[promise_pure, will_not_call_mercury],
"{
	double val;
	if (CLPR_is_ground(Svar, &val)) {
		printf(""%f"", val);
	} else {
		printf(""_v%ld"", (long) Svar);
	}
}").

	% WARNING: The implementation of this predicate is non-logical.
:- pragma promise_pure(dump_one_solution/3).

dump_one_solution(Pred, !IO) :-
	( Pred(Vars, VarNames) ->
		impure unsafe_dump(Vars, VarNames)
	;
		io.write_string("No solution.\n", !IO)
	).

	% Unfortunately the standard same_length predicate in list.m
	% doesn't have this mode...
	%
:- pred same_len(list(T1)::list_ca, list(T2)::in) is semidet.

same_len([], []).
same_len([_|Xs], [_|Ys]) :- same_len(Xs, Ys).

	% unsafe_dump(CfloatList, NameList) writes the constraint
	% relationships between the cfloats in the list CfloatList to
	% stdout, using the names given in the list NamesList.
	% Perhaps these should be one assoc_list(?).
	%
unsafe_dump(Cfloats, Names) :-
	% Ensure Cfloats and names have the same length
	( same_len(Cfloats, Names) ->
			% convert the mercury list of cfloats to a CLP(R)
			% representation.
		dump.mercury_cfloat_list_to_clpr_list(Cfloats, CLPRVars),
			% convert the mercury list of strings to the required
			% CLP(R) representation
		dump.mercury_string_list_to_clpr_dump_string(Names, CLPRNames),
		impure dump.do_dump(CLPRVars, CLPRNames),
		impure dump.free(CLPRVars, CLPRNames)
	;
		error("list length mismatch in dump")
	).

	% convert the mercury cfloat list to a CLP(R) representation.
	% The required representation is a CLPR_tagged pointer;
	% CLPR_NIL is the empty list, a CLPR_CONS CLPR_tagged pointer
	% points to a two cell area of memory, the first element of
	% which is the cfloat, the second of which if either NIL or
	% CONS (etc.). The cfloat itself is a pointer to a single
	% cell.  This cell has the CLPR_tag PAR, and the value (with
	% the CLPR_tag taken off) is the variable's CLPR_solver_id.
	%
:- pred dump.mercury_cfloat_list_to_clpr_list(list(cfloat)::list_ca, 
	c_pointer::out) is det.

dump.mercury_cfloat_list_to_clpr_list([], Empty) :- 
	dump.get_clpr_empty_list(Empty).
dump.mercury_cfloat_list_to_clpr_list([V|Vs], Result) :-
	dump.mercury_cfloat_list_to_clpr_list(Vs, Tail),
	dump.make_clpr_cfloat_list(V, Tail, Result).

:- pragma foreign_decl("C", "#include <stdio.h>").
:- pragma foreign_decl("C", "#include ""clpr/emul.h""").
:- pragma foreign_decl("C", "#include ""clpr/compile.h""").

:- pred dump.get_clpr_empty_list(c_pointer::out) is det.
:- pragma foreign_proc("C",
	dump.get_clpr_empty_list(Empty::out),
	[promise_pure, will_not_call_mercury],
"
	Empty = addtag(CLPR_TAG_NIL, 0);
").

:- pred dump.make_clpr_cfloat_list(cfloat::ca, c_pointer::in, c_pointer::out) 
	is det.

:- pragma foreign_proc("C",
	dump.make_clpr_cfloat_list(Head::ca, Tail::in, TheList::out),
	[promise_pure, will_not_call_mercury],
"{
	CLPR_int *HeadPtr;
	CLPR_int **ListPtr;

	ListPtr = malloc(2 * sizeof(CLPR_int *));
	HeadPtr = malloc(sizeof(CLPR_int));
	if (ListPtr == NULL || HeadPtr == NULL) {
		MR_fatal_error(""malloc() failed in dump"");
	}

	*HeadPtr = addtag(TAG_PAR, Head);
	ListPtr[0] = HeadPtr;
	ListPtr[1] = (CLPR_int *) Tail;
	TheList = addtag(CLPR_TAG_CONS, ListPtr);
}").

	% Convert the Mercury list of strings to the required CLP(R)
	% representation - a single string, with the individual strings
	% separated by the value DUMP_SEPARATOR.
	%
:-pred dump.mercury_string_list_to_clpr_dump_string(list(string)::in, 
	c_pointer::out) is det.

dump.mercury_string_list_to_clpr_dump_string(Strings, CLPRString) :-
	dump.total_string_length(Strings, StringsLength),
	list.length(Strings, NumStrings),
	CLPRStringLength = StringsLength + NumStrings,
	dump.allocate_clpr_string(CLPRStringLength, CLPRString0),
	dump.add_strings(Strings, CLPRString0, 0, CLPRString).

:- pred dump.total_string_length(list(string)::in, int::out) is det.

dump.total_string_length([], 0).
dump.total_string_length([X | Xs], Result) :-
	string.length(X, Length),
	dump.total_string_length(Xs, TheRest),
	Result = Length + TheRest.

:- pred dump.allocate_clpr_string(int::in, c_pointer::out) is det.
:- pragma foreign_proc("C",
	dump.allocate_clpr_string(Length::in, CLPRString::out),
	[promise_pure, will_not_call_mercury],
"{
	char *s;
	if ((s = malloc(Length * sizeof(char))) == NULL) {
		MR_fatal_error(
			""malloc() failed in dump.allocate_clpr_string"");
	}
	CLPRString = (MR_Word) s;
}").

:- pred dump.add_strings(list(string)::in, c_pointer::in, int::in, 
	c_pointer::out) is det.

dump.add_strings([], CLPRString, _, CLPRString).
dump.add_strings([X | Xs], CLPRString0, Index, CLPRString) :-
	dump.add_single_string(X, CLPRString0, Index, CLPRString1),
	string.length(X, XLength),
	NewIndex = Index + XLength + 1,
	dump.add_strings(Xs, CLPRString1, NewIndex, CLPRString).

:- pred dump.add_single_string(string::in, c_pointer::in, int::in, 
	c_pointer::out) is det.
:- pragma foreign_proc("C",
	dump.add_single_string(TheString::in, CLPRString0::in, 
		Index::in, CLPRString::out),
	[promise_pure, will_not_call_mercury],
"{
	char *CLPRTmp, *StringTmp;

	CLPRTmp = (char *) CLPRString0;
	StringTmp = (char *) TheString;

	CLPRString = CLPRString0;
	CLPRTmp += Index;
	while (*StringTmp != '\\0') {
		*CLPRTmp++ = *StringTmp++;
	}
	
	*CLPRTmp = DUMP_SEPARATOR;
}").

:- pragma foreign_decl("C", "extern void CLPR_dump1(FILE *, CLPR_int *, char *);").

:- impure pred dump.do_dump(c_pointer::in, c_pointer::in) is det.
:- pragma foreign_proc("C",
	dump.do_dump(Vars::in, Names::in),
	[will_not_call_mercury],
"
	CLPR_dump1(stdout, (CLPR_int *)&Vars, (char *)Names);
").

:- pragma foreign_decl("C", "static void free_clpr_list_mem(MR_Word list);").
:- pragma foreign_code("C", "
	static void free_clpr_list_mem(MR_Word list)
	{
		if (CLPR_tag(list) == CLPR_TAG_CONS) {
			MR_Word *list_ptr = (MR_Word *) pmask(list);
			free_clpr_list_mem(list_ptr[1]);
			free(list_ptr);
		}

	}
").

	% Free our dynamically allocated memory.
	%
:- impure pred dump.free(c_pointer::in, c_pointer::in) is det.
:- pragma foreign_proc("C",
	dump.free(Vars::in, TheString::in),
	[will_not_call_mercury],
"
	free_clpr_list_mem(Vars);
	free((char *) TheString);
").

:- pragma foreign_decl("C", "extern void CLPR_print_tableaus(void);").
:- pragma foreign_proc("C",
	unsafe_dump_tableaus,
	[may_call_mercury],
"
	CLPR_print_tableaus();
").

%----------------------------------------------------------------------------%
:- end_module dump.
%----------------------------------------------------------------------------%
