% tranny - a circuit constraint solver. 
%
% Author - Dr. Peter Stuckey (pjs@cs.mu.oz.au), converted to Mercury by 
%		David Jeffery (dgj@students.cs.mu.oz.au), and hacked a little
%		further by Fergus Henderson (fjh@cs.mu.oz.au).
%
% The conversion to Mercury is not complete - dc_analysis works correctly,
% so the top-level goal go1 works fine, but full_analysis has a bug.


:- module tranny.

:- interface.

:- import_module io.

:- pred main(io__state::di, io__state::uo) is cc_multi.

:- implementation.

:- import_module cfloat, float_cfloat, cfloat_float, float.
:- import_module require, std_util, int, list, dump.

:- type circuit		==	list(pair(component, list(node))).

:- inst circuit		==	list(pair(component, ground)).

:- type component	--->	resistor(resistor_name, resistor_value) 
			;	capacitor(capacitor_name)
			;	diode(diode_name, diode_data)
			;	transistor(transistor_name, transistor_info).

:- inst component
	--->	resistor(ground, any) 
	;	capacitor(ground)
	;	diode(ground, diode_data)
	;	transistor(ground, transistor_info).

:- type node 		--->	a ; b ; c ; d ; e ; p1 ;
				cc1 ; cc2 ; in ; gnd ; out.

:- type circuit_node	--->	n(node, cfloat, list(cfloat)).
				% Node name, Voltate, Collector Currents
:- inst circuit_node	==	bound(n(ground, any, list_skel(any))).

:- type resistor_name 	--->	r1 ; r2 ; r3 ; r4 ; re ; rc.
:- type capacitor_name 	--->	c1 ; c2 ; c3 ; c4.
:- type diode_name 	--->	d1 ; d2 ; d3 ; d4.
:- type transistor_name --->	t1 ; t2 ; t3 ; t4.

:- type resistor_value 	== 	cfloat.

	% XXX We could move this into the component type, for a *minor* 
	% efficiency gain.
:- type diode_data	--->	diode_data(diode_code, diode_state,
							cfloat, cfloat).
				% Diode code, Diode state, Vf, Vbreak.
:- inst diode_data == bound(	diode_data(ground, ground, any, any)).

:- type diode_code 	--->	di1.
:- type diode_state 	--->	forward ; reverse.

:- type transistor_info	--->	info(
					transistor_type, transistor_code, 
					transistor_state, transistor_data, 
					transistor_data, transistor_data
				).
				% Type, Code, State, Mean, Min, Max.
:- inst transistor_info == bound(info(
					ground, ground, ground,
					transistor_data,
					transistor_data,
					transistor_data
				)).
:- type transistor_data	--->	data(cfloat, cfloat, cfloat, cfloat).
				% Beta, Vbe, Vcestat, Vt).
:- inst transistor_data	== bound(data(any, any, any, any)).

:- type transistor_type --->	npn ; pnp.
:- type transistor_code --->	tr0 ; tr1.
:- type transistor_state --->	active ; saturated ; cutoff.

:- type comp_type 	--->	mean ; min ; max.

/****************************************************************************/ 
/*  Major goals                                                             */ 
/****************************************************************************/ 
:- pred dc_analysis(cfloat::ca, cfloat::ca, circuit::in(circuit),
	list(circuit_node)::out(list_skel(circuit_node))) is cc_nondet.
dc_analysis(Vcc1,Vcc2,Circuit, Nodelist) :-
	choose_circuit(Circuit), 
	cfloat__init(A),
	cfloat__init(B),
	cfloat__init(C),
	X == 0.0,
	Nodelist0 = [n(cc1,Vcc1,[A]),n(cc2,Vcc2,[B]),n(gnd,X,[C])], 
	solve_dc(mean, Circuit, Nodelist0, Nodelist, _Collector_Currents), 
	current_solve(Nodelist).

:- pred full_analysis(cfloat, cfloat, circuit, node, node, configuration,
	cfloat, cfloat, cfloat, cfloat).
/*
:- mode full_analysis(ca, ca, in(circuit), out, out, out, co, co, co, co)
	is nondet.
*/
:- mode full_analysis(ca, ca, out(circuit), out, out, out, co, co, co, co)
	is nondet.
full_analysis(Vcc1,Vcc2,Circuit,In,Out,Type,Stability,Gain,Inresist,Outresist):- 
%*** Choose a circuit template ***% 
	circuit(Vcc1, Vcc2, Circuit,In, Out,Type), 
%*** Construct constraints that hold over the circuit ***% 
	Zero == 0.0,
	cfloat__init(Z1), cfloat__init(Z2), cfloat__init(Z3),
	solve_dc(mean,Circuit ,
		[n(cc1,Vcc1,[Z1]), n(cc2,Vcc2,[Z2]), n(gnd,Zero,[Z3])], 
		Nodelist,Collector_Currents), 
	current_solve(Nodelist), 
 
%*** Determine stability constraints ****% 
	stability(Vcc1,Vcc2,Circuit,Collector_Currents,Stability), 
 
%*** Determine Input resistance and Gain constraints ***% 
	One == 1.0,
	cfloat__init(X1), cfloat__init(X2), cfloat__init(X3),
	cfloat__init(Iin),
	cfloat__init(Vout),
	SS = [n(cc1,Zero,[X1]),n(cc2,Zero,[X2]),n(gnd,Zero,[X3]), 
		 n(In,One,[Iin]),n(Out,Vout,[])],
	solve_ss(Circuit, Collector_Currents, SS, Nodelist2),
	current_solve(Nodelist2), 
	Inresist * Iin == -1.0,
	Gain = Vout, 
 
%*** Determine Output resistance constraints ****% 
	cfloat__init(Y1), cfloat__init(Y2), cfloat__init(Y3),
	cfloat__init(Iout),
	SS2 = [n(cc1,Zero,[Y1]),n(cc2,Zero,[Y2]),n(gnd,Zero,[Y3]), 
		 n(Out,One,[Iout])],
	solve_ss(Circuit,Collector_Currents, SS2, Nodelist3),
	current_solve(Nodelist3), 
	Iout * Outresist == -1.0,
 
%*** Choose circuit values - all (real) choice points occur here ***% 
	choose_circuit(Circuit). 
	 
/****************************************************************************/ 
/*  small signal equivalent solve                                          */ 
/****************************************************************************/ 
:- pred solve_ss(circuit, list(cfloat), list(circuit_node), list(circuit_node)).
:- mode solve_ss(in(circuit), list_ca, in(list_skel(circuit_node)),
		out(list_skel(circuit_node))) is semidet.
solve_ss([], [], List, List).
solve_ss([], [_|_], _, _) :-
	error("list length mismatch in solve_ss/4").
solve_ss([(Component - Points)|Rest], CCin, Innodes, Outnodes):- 
	connecting(Points, Volts, Amps, Innodes, Tmpnodes), 
	component_ss(Component, Volts, Amps, CCin, CCout),
	solve_ss(Rest, CCout, Tmpnodes, Outnodes).
 
:- pred component_ss(component, list(cfloat), list(cfloat), list(cfloat),
			list(cfloat)).
:- mode component_ss(in(component), list_ca, list_ca, list_ca, list_co)
	is semidet.
component_ss(resistor(_,R),[V1,V2],[I,-I],Cc,Cc):- 
	V1 - V2 = R * I.
component_ss(capacitor(_),[V,V],[I,-I],Cc,Cc).
component_ss(transistor(_,info(npn, _Code, active, Mean, _, _)), 
			[Vb,_Vc,Ve], [Ib,Ic,Ie],[Icol|CC],CC):- 
	Mean = data(Beta,_,_,Vt), 
	cfloat__init(TMP), % resolve type ambiguity
	Vt = TMP * Icol,
	Vb - Ve = (Beta * TMP) * Ib, 
	Ic = Beta * Ib, 
	Ie + Ic = -Ib. 

	 
/****************************************************************************/ 
/*  dc component solving                                                    */ 
/****************************************************************************/ 

:- pred solve_dc(comp_type, circuit, list(circuit_node), list(circuit_node),
	list(cfloat)).
:- mode solve_dc(in, in(circuit), in(list_skel(circuit_node)),
		out(list_skel(circuit_node)), list_co) is semidet.

	% XXX This could be more efficient if we used a functor instead of a
	% list for Volts etc.
solve_dc(_, [], List, List, []).
solve_dc(Kind,[(Component - Points) | Rest], Inlist, Outlist, CC):- 
	connecting(Points, Volts, Amps, Inlist,Tmplist), 
	solve_dc(Kind, Rest, Tmplist, Outlist, CC0),
	component_dc(Component, Volts, Amps, CC0, CC, Kind).
 
	% component_type(Component, Volts, Amps, CCin, CCout, ComputationType).
:- pred component_dc(component, list(cfloat), list(cfloat), list(cfloat),
	list(cfloat), comp_type).
:- mode component_dc(in(component), list_ca, list_ca, list_ca, list_co, in)
	is semidet.

component_dc(resistor(_, R), [V1, V2], [I, -I], Cc,Cc,_) :- 
	V1 - V2 = R * I.
component_dc(capacitor(_), _, [V1, V2], Cc, Cc, _) :-
	V1 == 0.0, V2 == 0.0.
component_dc(transistor(_, Info), [V1, V2, V3], [Ib, Ic, Ie], CC0, CC,
		CompType):- 
	Info = info(Type, _Code, State, Mean, Min, Max), 
	(
		CompType = mean,
		Mean = data(Beta, Vbe, Vcesat, _)
	;
		CompType = min,
		Min = data(Beta, Vbe, Vcesat, _)
	;
		CompType = max,
		Max = data(Beta, Vbe, Vcesat, _)
	),
	transistor_state(Type,State,Beta,Vbe,Vcesat, V1, V2, V3, Ib, Ic, Ie),
	CC = [Ic | CC0].
component_dc(diode(_Code, Data), [V1, V2], [I1, I2], Cc, Cc, _):- 
	Data = diode_data(_, State, Vf, Vbreak),
	diode_state(State, Vf, Vbreak, V1, V2, I1, I2).

/****************************************************************************/ 
/*   diode and transistor states / and relationships                        */ 
/****************************************************************************/ 
:- pred diode_state(diode_state, cfloat, cfloat, cfloat, cfloat, cfloat,
		cfloat).
:- mode diode_state(in, ca, ca, ca, ca, ca, ca) is semidet.

diode_state(forward, Vf, _Vbreak, Vp, Vm, I, -I):- 
	/* forward biased */ 
	Vp - Vm = Vf,
	I >= 0.0.

diode_state(reverse, Vf, Vbreak, Vp, Vm, I, -I):- 
	/* reverse biased */ 
	Vp - Vm < Vf,
	Vm - Vp < Vbreak,
	I == 0.0. 

:- pred transistor_state(transistor_type, transistor_state,
	cfloat, cfloat, cfloat, 
	cfloat, cfloat, cfloat,	% The voltages
	cfloat, cfloat, cfloat).	% The currents
:- mode transistor_state(in, in, ca, ca, ca, ca, ca, ca, ca, ca, ca) is semidet.

transistor_state(npn, active, Beta, Vbe, _, Vb, Vc, Ve, Ib, Ic, Ie):- 
	Vb = Ve + Vbe, 
	Vc >= Vb,
	Ib >= 0.0,
	Ic = Beta * Ib, 
	Ie + Ib = -Ic.
transistor_state(pnp, active, Beta, Vbe,_, Vb, Vc, Ve, Ib, Ic, Ie):- 
	Vb = Ve + Vbe, 
	Vc =< Vb,
	Ib =< 0.0,
	Ic = Beta * Ib, 
	Ie + Ib = -Ic.
transistor_state(npn, saturated, _Beta, Vbe, Vcesat, Vb, Vc, Ve, Ib, Ic, Ie):-
	Vb = Ve + Vbe, 
	Vc = Ve + Vcesat, 
	Ib >= 0.0,
	Ic >= 0.0,
	Ie + Ib = -Ic.
transistor_state(pnp, saturated, _Beta, Vbe, Vcesat, Vb, Vc, Ve, Ib, Ic, Ie):-
	Vb = Ve + Vbe, 
	Vc = Ve + Vcesat, 
	Ib =< 0.0,
	Ic =< 0.0,
	Ie + Ib = -Ic.
transistor_state(npn, cutoff, _Beta, Vbe, _Vcesat, Vb, _Vc, Ve, Ib, Ic, Ie):- 
	Vb =< Ve + Vbe, 
	Ib == 0.0, 
	Ic == 0.0, 
	Ie == 0.0. 
transistor_state(pnp, cutoff, _Beta, Vbe, _Vcesat, Vb, _Vc, Ve, Ib, Ic, Ie):- 
	Vb >= Ve + Vbe, 
	Ib == 0.0, 
	Ic == 0.0, 
	Ie == 0.0. 
 
/****************************************************************************/ 
/*   connecting components routines                                         */ 
/****************************************************************************/ 
:- pred connecting(list(node), list(cfloat), list(cfloat), list(circuit_node), 
	list(circuit_node)).
:- mode connecting(in, list_co, list_co, in(list_skel(circuit_node)),
			out(list_skel(circuit_node))) is det.
connecting([], [], [], List, List). 
connecting([P|PR], [V|VR], [I|IR], Inlist, Outlist):- 
	connect(P, V, I, Inlist, Tmplist), 
	connecting(PR, VR, IR, Tmplist, Outlist).

:- pred connect(node, cfloat, cfloat, list(circuit_node), list(circuit_node)).
:- mode connect(in, co, co, in(list_skel(circuit_node)),
		out(list_skel(circuit_node))) is det.
connect(P, V, I, [], [n(P, V, [I])]) :- cfloat__init(V), cfloat__init(I).
connect(P, V, I, [N|Ns], Result) :-
	(
		N = n(P, V1, Ilist)
	->
		V = V1,
		cfloat__init(I),
		Result = [n(P, V, [I|Ilist])|Ns]
	;
		connect(P, V, I, Ns, Newrest), 
		Result = [N|Newrest]
	).

/****************************************************************************/
/*  Stability Analysis                                                      */
/****************************************************************************/
:- pred stability(cfloat, cfloat, circuit, list(cfloat), cfloat).
:- mode stability(ca, ca, in(circuit), list_ca, co) is semidet.
stability(Vcc1,Vcc2,Circuit, CollectorCurrents, Stability):-
	Zero == 0.0,
	cfloat__init(X1), cfloat__init(X2), cfloat__init(X3), 
	cfloat__init(X4), cfloat__init(X5), cfloat__init(X6),
        solve_dc(min, Circuit,
		[n(cc1,Vcc1,[X1]),n(cc2,Vcc2,[X2]),n(gnd,Zero,[X3])],
                 Nodelist1,MinCurrents),
        current_solve(Nodelist1),
        solve_dc(max, Circuit,
		[n(cc1,Vcc1,[X4]),n(cc2,Vcc2,[X5]),n(gnd,Zero,[X6])],
                 Nodelist2,MaxCurrents),
        current_solve(Nodelist2),
        calculate(MinCurrents,MaxCurrents,CollectorCurrents,Stability).

:- pred calculate(list(cfloat), list(cfloat), list(cfloat), cfloat).
:- mode calculate(list_ca, list_ca, list_ca, co) is semidet.
calculate(MinCurrents,MaxCurrents,CollectorCurrents,Stability):-
        cal(MinCurrents,MaxCurrents,CollectorCurrents,Percents),
	Zero == 0.0,
        maxi(Percents,Zero,Stability).

:- pred cal(list(cfloat), list(cfloat), list(cfloat), list(cfloat)).
:- mode cal(list_ca, list_ca, list_ca, list_co) is semidet.
cal([], [], [], []).
cal([Min|Rin], [Max|Rax], [Ic|Rc], [Pc|Rpc]) :-
	Pc = max(Ic - Min, Max - Ic),
	cal(Rin, Rax, Rc, Rpc).

:- pred maxi(list(cfloat), cfloat, cfloat).
:- mode maxi(list_ca, ca, co) is det.
maxi([N1|R],N2,P):-
        M = max(N1,N2),
        maxi(R,M,P).
maxi([],P,P).

/****************************************************************************/ 
/*  Miscellaneous things                                                    */ 
/****************************************************************************/ 
:- pred current_solve(list(circuit_node)).
:- mode current_solve(in(list_skel(circuit_node))) is semidet.
current_solve([]). 
current_solve([n(_,_,L) | Rest]) :- 
	kcl(L), 
	current_solve(Rest).
 
:- pred sum(list(cfloat), cfloat).
:- mode sum(list_ca, co) is det.
sum([], Z) :-
	Z == 0.0.
sum([X|T], Z) :- 
	sum(T, P),
	Z = X + P.
 
:- pred kcl(list(cfloat)).
:- mode kcl(list_ca) is semidet.
kcl(L) :-  
	sum(L, X),
	X == 0.0.
 
:- pred print_value(list(circuit_node)::in(list_skel(circuit_node)),
		io__state::di, io__state::uo) is cc_multi.
print_value([]) --> []. 
print_value([n(P,V,I) | Rest]) --> 
	io__write(P),
	io__write_string(" at "),
	dump_cfloat(V),
	io__write_string(" "),
	dump_cfloat_list(I),
	io__write_string("\n"),
	print_value(Rest). 

:- pred dump_cfloat_list(list(cfloat)::list_ca, 
	io__state::di, io__state::uo) is cc_multi.
dump_cfloat_list(X) -->
	io__write_string("["),
	dump_cfloat_list_2(X),
	io__write_string("]").

:- pred dump_cfloat_list_2(list(cfloat)::list_ca, 
	io__state::di, io__state::uo) is cc_multi.
dump_cfloat_list_2([]) --> [].
dump_cfloat_list_2([X|Xs]) -->
	dump_cfloat(X),
	( { Xs \= [] } ->
		io__write_string(", ")
	;	
		[]
	),
	dump_cfloat_list_2(Xs).

:- pred print_circuit(circuit::in(circuit), io__state::di, io__state::uo)
	is cc_multi.
print_circuit([]) --> []. 
print_circuit([(Component - _Points) | Rest]) -->
	print_component(Component),
	print_circuit(Rest). 

:- pred print_component(component::in(component),
		io__state::di, io__state::uo) is cc_multi.
print_component(resistor(Name, Value)) -->
	io__write_string("resistor at "),
	io__print(Name),
	io__write_string(" "),
	dump_cfloat(Value),
	io__write_string("\n").
print_component(capacitor(Name)) -->
	io__write_string("capacitor at "),
	io__print(Name),
	io__write_string("\n").
print_component(diode(Name, Data)) -->
	io__write_string("diode at "),
	io__print(Name),
	io__write_string(" "),
	write_diode_data(Data),
	io__write_string("\n").
print_component(transistor(Name, Info)) -->
	io__write_string("transistor at "),
	io__print(Name),
	io__write_string(" "),
	write_transistor_info(Info),
	io__write_string("\n").

:- pred write_diode_data(diode_data, io__state, io__state).
:- mode write_diode_data(in(diode_data), di, uo) is cc_multi.

write_diode_data(diode_data(Code, State, Vf, Vbreak)) -->
	print("["),
	print(Code),
	print(", "),
	print(State),
	print(", "),
	dump_cfloat(Vf),
	print(", "),
	dump_cfloat(Vbreak),
	print("]").

:- pred write_transistor_info(transistor_info, io__state, io__state).
:- mode write_transistor_info(in(transistor_info), di, uo) is cc_multi.

write_transistor_info(info(Type, Code, State, DataMean, DataMin, DataMax)) -->
	print("["),
	print(Type),
	print(", "),
	print(Code),
	print(", "),
	print(State),
	print(", "),
	write_transistor_data(DataMean),
	print(", "),
	write_transistor_data(DataMin),
	print(", "),
	write_transistor_data(DataMax),
	print("]").

:- pred write_transistor_data(transistor_data, io__state, io__state).
:- mode write_transistor_data(in(transistor_data), di, uo) is cc_multi.

write_transistor_data(data(Beta, Vbe, Vcestat, Vt)) -->
	print("info("),
	dump_cfloat(Beta),
	print(", "),
	dump_cfloat(Vbe),
	print(", "),
	dump_cfloat(Vcestat),
	print(", "),
	dump_cfloat(Vt),
	print(")").

:- pred print_amplifier_type(configuration::in, io__state::di, io__state::uo)
	is det.
print_amplifier_type(common_emitter) -->
	io__write_string("Amplifier type: common emitter\n").
print_amplifier_type(common_base) -->
	io__write_string("Amplifier type: common base\n").
print_amplifier_type(emitter_follower) -->
	io__write_string("Amplifier type: emitter follower\n").

/****************************************************************************/ 
/* Choose circuit values                                                    */ 
/****************************************************************************/ 
:- pred choose_circuit(circuit).
:- mode choose_circuit(in(circuit)) is nondet.
choose_circuit([(Component - _ ) | RestofCircuit]):- 
	choose_component(Component), 
	choose_circuit(RestofCircuit).
choose_circuit([]). 

 
:- pred choose_component(component).
:- mode choose_component(in(component)) is nondet.
choose_component(resistor(_, R)) :-
	Cfloat_R == R,
	resistor_val(Cfloat_R).  
choose_component(capacitor(_)). 
choose_component(diode(_, diode_data(Code,_,Vf,Vbreak))):- 
	diode_type(Code,Vf,Vbreak).
choose_component(transistor(_, info(Type,Code,_,Mean,Min,Max))):- 
	transistor_type(Type,Code, MeanBeta, MeanVbe, MeanVcestat, MeanVt, 
		mean), 
	Mean = data(MeanBeta,MeanVbe,MeanVcestat,MeanVt), 
	transistor_type(Type,Code, MinBeta, MinVbe, MinVcestat, MinVt, min), 
	Min = data(MinBeta,MinVbe,MinVcestat,MinVt), 
	transistor_type(Type,Code, MaxBeta, MaxVbe, MaxVcestat, MaxVt, max), 
	Max = data(MaxBeta,MaxVbe,MaxVcestat,MaxVt).
/****************************************************************************/ 
/* Database of circuits and components                                      */ 
/****************************************************************************/ 
:- pred resistor_val(cfloat).
:- mode resistor_val(ca) is nondet.

resistor_val(X) :-
	( X == 100.0
	; X == 50.0
	; X == 5.0
	; X == 1.0
	).
 
:- pred diode_type(diode_code, cfloat, cfloat).
:- mode diode_type(in, ca, ca) is semidet.
diode_type(di1, X, Y) :-
	X == 0.6,
	Y == 100.00. 

:- pred transistor_type(transistor_type, transistor_code, 
	cfloat, cfloat, cfloat, cfloat, comp_type).
:- mode transistor_type(in, in, ca, ca, ca, ca, in) is semidet.

transistor_type(npn, tr0, W, X, Y, Z, mean) :- 
	W == 100.0,
	X == 0.7,
	Y == 0.3,
	Z == 0.025.
transistor_type(npn, tr0, W, X, Y, Z, min) :- 
	W == 50.0,
	X == 0.8,
	Y == 0.3,
	Z == 0.025.
transistor_type(npn, tr0, W, X, Y, Z, max) :- 
	W == 150.0,
	X == 0.6,
	Y == 0.3,
	Z == 0.025.
transistor_type(pnp, tr1, W, X, Y, Z, mean) :-
	W == 100.0,
	X == -0.7,
	Y == -0.3,
	Z == 0.025.
transistor_type(pnp, tr1, W, X, Y, Z, min) :-
	W == 50.0,
	X == -0.8,
	Y == -0.3,
	Z == 0.025.
transistor_type(pnp, tr1, W, X, Y, Z, max) :-
	W == 150.0,
	X == -0.6,
	Y == -0.3,
	Z == 0.025.
 
 :- type configuration	--->	common_emitter
 			;	common_base
			;	emitter_follower.

:- pred circuit(cfloat::ca, cfloat::ca, circuit::out(circuit), node::out,
	node::out, configuration::out) is nondet.
circuit(X, Y, Circuit, in, out, common_emitter) :- 
	X == 15.0,
	Y == 0.0,
	cfloat__init(R1), cfloat__init(R2),
	cfloat__init(Re), cfloat__init(Rc),
	transistor_data(Mean),
	transistor_data(Minn),
	transistor_data(Maxx),
	Circuit = [
		capacitor(c1) - [in,b], 
		resistor(r1,R1) - [b,cc1], 
		resistor(r2,R2) - [b,gnd], 
		transistor(t1,info(npn,tr0,active,Mean,Minn,Maxx)) - [b,c,e],
		resistor(re,Re) - [e,gnd],
		capacitor(c2) - [c,out],
		resistor(rc,Rc) - [c,cc1], 
		capacitor(c3) -[e,gnd]
		].
 
circuit(X, Y, Circuit, in, out, common_base) :- 
	X == 15.0,
	Y == 0.0,
	cfloat__init(R1), cfloat__init(R2), cfloat__init(Re), cfloat__init(Rc),
	transistor_data(Mean),
	transistor_data(Minn),
	transistor_data(Maxx),
	Circuit = [
		capacitor(c1) - [gnd,b], 
		resistor(r1,R1) - [b,cc1], 
		resistor(r2,R2) - [b,gnd], 
		transistor(t2, info(pnp,tr1,active,Mean,Minn,Maxx)) - [b,c,e], 
		resistor(re,Re) - [e,gnd], 
		capacitor(c2) - [c,in], 
		resistor(rc,Rc) - [c,cc1], 
		capacitor(c3) - [e,out]
		].
 
circuit(X, Y, Circuit, in, out, emitter_follower) :- 
	X == 15.0,
	Y == 0.0,
	cfloat__init(R1), cfloat__init(R2), cfloat__init(Re),
	transistor_data(Mean),
	transistor_data(Minn),
	transistor_data(Maxx),
	Circuit = [
		capacitor(c1) - [in,b], 
		resistor(r1, R1) - [b,cc1], 
		resistor(r2, R2) - [b,gnd], 
		transistor(t3,info(npn,tr0,active,Mean,Minn,Maxx)) - [b,cc1,e], 
		resistor(re,Re) - [e,gnd], 
		capacitor(c3) - [e,out] 
		].
 
:- pred transistor_data(transistor_data).
:- mode transistor_data(out(transistor_data)) is det.
transistor_data(data(X1, X2, X3, X4)) :-
	cfloat__init(X1), cfloat__init(X2), cfloat__init(X3), cfloat__init(X4).

:- pred transistor_state(transistor_state).
:- mode transistor_state(in) is det.
:- mode transistor_state(out) is multi.
transistor_state(active).
transistor_state(saturated).
transistor_state(cutoff).

:- pred diode_state(diode_state).
:- mode diode_state(in) is det.
:- mode diode_state(out) is multi.
diode_state(forward).
diode_state(reverse).
 
:- pred go1(io__state::di, io__state::uo) is cc_multi.
go1 -->
	(
		{
		R1 == 100.0,
		R2 == 100.0,
		R3 == 5.0,
		transistor_data(Data1),
		transistor_data(Data2),
		transistor_data(Data3),
		transistor_state(State),
		Circuit = [ 
				(capacitor(c1) - [in,b]), 
				(resistor(r1,R1) - [b,cc1]), 
				(resistor(r2,R2) - [b,gnd]), 
				(transistor(t1,info(npn,tr0,State,
					Data1,Data2,Data3)) - [b,cc1,e]), 
				(resistor(r3,R3) - [e,gnd]), 
				(capacitor(c3) - [e,out])
			],
		VCC1 == 15.0,
		cfloat__init(VCC2),
		dc_analysis(VCC1, VCC2, Circuit, Nodelist)
		}
	->
		print_circuit(Circuit),
		print_value(Nodelist),
		print("Transistor state = "), print(State), nl
	;
		print("No solution\n")
	).

:- pred go2(io__state::di, io__state::uo) is cc_multi.
go2 -->
	(
		{
		transistor_data(TD1),
		transistor_data(TD2),
		transistor_data(TD3),
		transistor_state(State),
		diode_state(St1),
		diode_state(St2),
		diode_state(St3),
		diode_state(St4),
		init(D1Vf), init(D1Vb),
		init(D2Vf), init(D2Vb),
		init(D3Vf), init(D3Vb),
		init(D4Vf), init(D4Vb),

		Input1 = gnd, Input2 = cc1,
		R1 == 100.0,
		R2 == 100.0,
		RC == 100.0,
		Circuit = [ 
			diode(d1,diode_data(di1,St1,D1Vf,D1Vb)) - [a,Input1],
			diode(d2,diode_data(di1,St2,D2Vf,D2Vb)) - [a,Input2],
			resistor(r1,R1) - [a,cc1],
			diode(d3,diode_data(di1,St3,D3Vf,D3Vb)) - [a,p1],
			diode(d4,diode_data(di1,St4,D4Vf,D4Vb)) - [p1,b],
			resistor(r2,R2) - [b,cc2], 
			transistor(t1,info(npn,tr0,State,TD1,TD2,TD3)) -
				[b,c,gnd], 
			resistor(rc,RC) - [c,cc1]
		],
		VCC1 == 15.0,
		VCC2 == -12.0,
		dc_analysis(VCC1, VCC2, Circuit, Nodelist)
		}
	->
		print_circuit(Circuit),
		print_value(Nodelist),
		print("Transistor state = "), print(State), nl
	;
		print("No solution\n")
	).

:- pred go3(io__state::di, io__state::uo) is cc_multi.
go3	-->
	(
		{ Vcc1 == 15.0 }, 
		{ cfloat__init(Vcc2) },
		{ full_analysis(Vcc1, Vcc2, Circuit, _, _, Amp_Type, Stability, 
			Gain, Inresistance, Outresistance) }, 
		{ Stability < 0.5 }, 
		{ Gain > 0.5 }, 
        	{ Inresistance >= 25.0 }, 
		{ Outresistance =< 0.5 }
	->
		print_amplifier_type(Amp_Type), 
		print_circuit(Circuit) 
	;
		print("No solution.\n")
	).
 

/*
go4:- Vcc = 15, Stability < 0.5, Gain > 0.5, 
        Inresistance >= 25, Outresistance <= 0.5, 
        new_analysis(Vcc, _, Circuit, _, _, Amp_Type, Stability, 
		Gain, Inresistance, Outresistance), 
		printf(" Amplifier Type %\n",[Amp_Type]), 
		print_circuit(Circuit). 
 

go5:- Vcc = 15, Stability < 0.5, Gain > 0.5, 
        Inresistance >= 25, Outresistance <= 0.5, 
		Circuit = [ 
		[capacitor,c1,C1,[in,b]], 
		[resistor,r1,R1,[b,cc1]], 
		[resistor,r2,R2,[b,gnd]], 
		[transistor,tr,[npn,tr0,active,Mean,Minn,Maxx],[b,cc1,e]], 
		[resistor,re,Re,[e,gnd]], 
		[capacitor,c3,C3,[e,out]]],
		R2 = 50,
        new_analysis(Vcc, _, Circuit, _, _, Amp_Type, Stability, 
		Gain, Inresistance, Outresistance), 
		printf(" Amplifier Type %\n",[Amp_Type]), 
		print_circuit(Circuit). 
 
go6:- Vcc = 15, Stability < 0.5, Gain > 0.5, 
        Inresistance >= 25, Outresistance <= 0.5, 
		Circuit = [ 
		[capacitor,c1,C1,[in,b]], 
		[resistor,r1,R1,[b,cc1]], 
		[resistor,r2,R2,[b,gnd]], 
		[transistor,tr,[npn,tr0,active,Mean,Minn,Maxx],[b,cc1,e]], 
		[resistor,re,Re,[e,gnd]], 
		[capacitor,c3,C3,[e,out]]],
		Re = 100,
        new_analysis(Vcc, _, Circuit, _, _, Amp_Type, Stability, 
		Gain, Inresistance, Outresistance), 
		printf(" Amplifier Type %\n",[Amp_Type]), 
		print_circuit(Circuit). 
 

go7:- Vcc = 15, Stability < 0.5, Gain > 0.5, 
        Inresistance >= 25, Outresistance <= 0.5, 
		Circuit = [ 
		[capacitor,c1,C1,[in,b]], 
		[resistor,r1,R1,[b,cc1]], 
		[resistor,r2,R2,[b,gnd]], 
		[transistor,tr,[npn,tr0,active,Mean,Minn,Maxx],[b,cc1,e]], 
		[resistor,re,Re,[e,gnd]], 
		[capacitor,c3,C3,[e,out]]],
		R2 = 50,
        full_analysis(Vcc, _, Circuit, _, _, Amp_Type, Stability, 
		Gain, Inresistance, Outresistance), 
		printf(" Amplifier Type %\n",[Amp_Type]), 
		print_circuit(Circuit). 

go8:- Vcc = 15, Stability < 0.5, Gain > 0.5, 
        Inresistance >= 25, Outresistance <= 0.5, 
		Circuit = [ 
		[capacitor,c1,C1,[in,b]], 
		[resistor,r1,R1,[b,cc1]], 
		[resistor,r2,R2,[b,gnd]], 
		[transistor,tr,[npn,tr0,active,Mean,Minn,Maxx],[b,cc1,e]], 
		[resistor,re,Re,[e,gnd]], 
		[capacitor,c3,C3,[e,out]]],
		Re = 100,
        full_analysis(Vcc, _, Circuit, _, _, Amp_Type, Stability, 
		Gain, Inresistance, Outresistance), 
		printf(" Amplifier Type %\n",[Amp_Type]), 
		print_circuit(Circuit). 

go9:- Vcc = 15, Stability < 0.5, Gain > 0.5, 
        Inresistance >= 25, Outresistance <= 0.5, 
		R1 = 100, R2 = 50, Re = 100,
		Circuit = [ 
		[capacitor,c1,C1,[in,b]], 
		[resistor,r1,R1,[b,cc1]], 
		[resistor,r2,R2,[b,gnd]], 
		[transistor,tr,[npn,tr0,active,Mean,Minn,Maxx],[b,cc1,e]], 
		[resistor,re,Re,[e,gnd]], 
		[capacitor,c3,C3,[e,out]]],
		R2 = 50,
        full_analysis(Vcc, _, Circuit, _, _, Amp_Type, Stability, 
		Gain, Inresistance, Outresistance), 
		printf(" Amplifier Type %\n",[Amp_Type]), 
		print_circuit(Circuit). 
*/

main -->
	print("go1."), nl, go1, nl,
	print("go2."), nl, go2, nl,
	print("go3."), nl, go3, nl.
